import logging
from typing import TypeVar

from aioinject import InjectionContext, Provider
from aioinject.extensions import OnResolveExtension


T = TypeVar("T")


class MyExtension(OnResolveExtension):
    async def on_resolve(
        self,
        context: InjectionContext,  # noqa: ARG002
        provider: Provider[T],
        instance: T,  # noqa: ARG002
    ) -> None:
        logging.info("%s type was provided!", provider.type_)



import pytest

from aioinject import Container, providers
from aioinject.validation import (
    all_providers_for_type_have_equal_lifetime,
    validate_container,
)
from aioinject.validation.error import ContainerValidationErrorGroup


_VALIDATORS = [all_providers_for_type_have_equal_lifetime]


class IDependency:
    pass


class SingletonImpl(IDependency):
    pass


class ScopedImpl(IDependency):
    pass


def test_ok() -> None:
    container = Container()
    container.register(providers.Scoped(ScopedImpl, type_=IDependency))
    container.register(providers.Scoped(SingletonImpl, type_=IDependency))

    validate_container(container, _VALIDATORS)


def test_err() -> None:
    container = Container()
    container.register(providers.Scoped(ScopedImpl, type_=IDependency))
    container.register(providers.Singleton(SingletonImpl, type_=IDependency))

    with pytest.raises(ContainerValidationErrorGroup) as exc_info:
        validate_container(container, _VALIDATORS)

    assert len(exc_info.value.errors) == 1
    err = exc_info.value.errors[0]
    assert "has providers with different scopes" in err.message



from collections.abc import Sequence
from typing import Any

import aioinject
from aioinject import Provider
from benchmark.dependencies import (
    RepositoryA,
    RepositoryB,
    ServiceA,
    ServiceB,
    UseCase,
    create_session,
)


providers: Sequence[Provider[Any]] = [
    aioinject.Scoped(create_session),
    aioinject.Scoped(RepositoryA),
    aioinject.Scoped(RepositoryB),
    aioinject.Scoped(ServiceA),
    aioinject.Scoped(ServiceB),
    aioinject.Scoped(UseCase),
]


def create_container() -> aioinject.Container:
    container = aioinject.Container()
    container.register(*providers)
    return container



import time
from collections.abc import AsyncIterator
from datetime import timedelta
from typing import Annotated

import httpx
from fastapi import APIRouter, Depends, FastAPI
from httpx import ASGITransport

from aioinject import Inject
from aioinject.ext.fastapi import AioInjectMiddleware, inject
from benchmark.benches._common import UseCaseDepends
from benchmark.container import create_container
from benchmark.dependencies import (
    RepositoryA,
    RepositoryB,
    ServiceA,
    ServiceB,
    UseCase,
    create_session,
)
from benchmark.dto import BenchmarkResult


router = APIRouter()


@router.get("/aioinject")
@inject
async def test_aioinject(use_case: Annotated[UseCase, Inject]) -> int:
    return await use_case.execute()


@router.get("/depends")
async def test_depends(use_case: Annotated[UseCase, Depends()]) -> int:
    return await use_case.execute()


@router.get("/depends-wrap-async")
async def test_depends_wrap_async(use_case: UseCaseDepends) -> int:
    return await use_case.execute()


@router.get("/by-hand")
async def test_by_hand() -> int:
    async with create_session() as session:
        use_case = UseCase(
            service_a=ServiceA(repository=RepositoryA(session=session)),
            service_b=ServiceB(repository=RepositoryB(session=session)),
        )

    return await use_case.execute()


async def fastapi_bench(
    iterations: int,
    endpoint: str,
    *,
    enable_aioinject: bool = False,
) -> AsyncIterator[BenchmarkResult]:
    app = FastAPI()
    if enable_aioinject:
        app.add_middleware(AioInjectMiddleware, container=create_container())

    app.include_router(router)

    durations = []
    async with httpx.AsyncClient(
        transport=ASGITransport(app),
        base_url="http://test",
    ) as client:
        for _ in range(iterations):
            start = time.perf_counter()
            response = await client.get(endpoint)
            durations.append(
                timedelta(seconds=time.perf_counter() - start),
            )
            response.raise_for_status()
            assert response.content == b"42"  # noqa: S101

    yield BenchmarkResult(
        name=f"FastAPI - {endpoint}",
        durations=durations,
        iterations=iterations,
    )



import dataclasses
from typing import TYPE_CHECKING, Annotated, Generic, TypeAlias, TypeVar


@dataclasses.dataclass(slots=True)
class Inject:
    pass


T = TypeVar("T")

if TYPE_CHECKING:
    Injected: TypeAlias = Annotated[T, Inject]

else:

    class Injected(Generic[T]):
        def __class_getitem__(cls, item: object) -> object:
            return Annotated[item, Inject]



from collections.abc import Awaitable, Callable
from typing import Any

from aiogram import BaseMiddleware, Router
from aiogram.types import TelegramObject

import aioinject
from aioinject import _utils, decorators
from aioinject._types import P, T


__all__ = ["inject", "AioInjectMiddleware"]


def inject(function: Callable[P, T]) -> Callable[P, T]:
    wrapper = decorators.inject(
        function,
        inject_method=decorators.InjectMethod.context,
    )
    return _utils.clear_wrapper(wrapper)


class AioInjectMiddleware(BaseMiddleware):
    def __init__(self, container: aioinject.Container) -> None:
        self.container = container

    async def __call__(
        self,
        handler: Callable[[TelegramObject, dict[str, Any]], Awaitable[Any]],
        event: TelegramObject,
        data: dict[str, Any],
    ) -> Any:
        async with self.container.context():
            return await handler(event, data)

    def add_to_router(self, router: Router) -> None:
        for observer in router.observers.values():
            observer.outer_middleware.register(middleware=self)



from __future__ import annotations

import contextvars
import inspect
from collections import defaultdict
from collections.abc import Callable, Coroutine, Iterable, Mapping, Sequence
from contextvars import ContextVar
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Literal,
    TypeVar,
    overload,
)

from typing_extensions import Self

from aioinject._features.generics import get_generic_parameter_map
from aioinject._store import InstanceStore, NotInCache
from aioinject._types import AnyCtx, T
from aioinject.extensions import (
    ContextExtension,
    OnResolveExtension,
    SyncContextExtension,
    SyncOnResolveExtension,
)
from aioinject.providers import Dependency, DependencyLifetime


if TYPE_CHECKING:
    from aioinject import Provider, _types
    from aioinject.containers import Container

_T = TypeVar("_T")
_TExtension = TypeVar("_TExtension")

context_var: ContextVar[AnyCtx] = ContextVar("aioinject_context")
container_var: ContextVar[Container] = ContextVar("aioinject_container")


class _BaseInjectionContext(Generic[_TExtension]):
    def __init__(
        self,
        container: Container,
        singletons: InstanceStore,
        extensions: Sequence[_TExtension],
    ) -> None:
        self._container = container
        self._extensions = extensions

        self._singletons = singletons
        self._store = InstanceStore()

        self._token: contextvars.Token[AnyCtx] | None = None
        self._providers: _types.Providers[Any] = defaultdict(list)

        self._closed = False

    def _get_store(self, lifetime: DependencyLifetime) -> InstanceStore:
        if lifetime is DependencyLifetime.singleton:
            return self._singletons
        return self._store

    def _get_providers(self, type_: type[_T]) -> list[Provider[_T]]:
        return self._providers.get(type_) or self._container.get_providers(
            type_,
        )

    def register(self, provider: Provider[Any]) -> None:
        self._providers[provider.type_].append(provider)


class InjectionContext(_BaseInjectionContext[ContextExtension]):
    async def resolve(self, type_: type[_T]) -> _T:
        return await self._resolve(type_, is_iterable=False)

    async def resolve_iterable(self, type_: type[_T]) -> list[_T]:
        return await self._resolve(type_, is_iterable=True)

    @overload
    async def _resolve(
        self,
        type_: type[_T],
        *,
        is_iterable: Literal[False],
    ) -> _T: ...

    @overload
    async def _resolve(
        self,
        type_: type[_T],
        *,
        is_iterable: Literal[True],
    ) -> list[_T]: ...

    async def _resolve(
        self,
        type_: type[_T],
        *,
        is_iterable: bool,
    ) -> _T | list[_T]:
        providers = self._get_providers(type_)
        if not is_iterable:
            return await self._resolve_provider(providers[-1])  # type: ignore[arg-type]
        return [
            await self._resolve_provider(provider) for provider in providers
        ]

    async def _resolve_provider(
        self,
        provider: Provider[_T],
    ) -> _T:
        store = self._get_store(provider.lifetime)
        if (cached := store.get(provider)) is not NotInCache.sentinel:
            return cached

        provider_dependencies = provider.collect_dependencies(
            context=self._container.type_context
        )
        dependencies_map = get_generic_parameter_map(
            provider.type_,  # type: ignore[arg-type]
            provider_dependencies,
        )
        dependencies = {
            dependency.name: await self._resolve(  # type: ignore[call-overload]
                type_=dependencies_map.get(
                    dependency.name,
                    dependency.inner_type,
                ),
                is_iterable=dependency.is_iterable,
            )
            for dependency in provider_dependencies
        }

        if provider.lifetime is DependencyLifetime.singleton:
            async with store.lock(provider) as should_provide:
                if should_provide:
                    return await self._provide_and_store(
                        provider, store, dependencies
                    )
                return store.get(provider)  # type: ignore[return-value] # pragma: no cover

        return await self._provide_and_store(provider, store, dependencies)

    async def _provide_and_store(
        self,
        provider: Provider[_T],
        store: InstanceStore,
        dependencies: Mapping[str, object],
    ) -> _T:
        provided = await provider.provide(dependencies)
        if provider.is_generator:
            provided = await store.enter_context(provided)
        store.add(provider, provided)
        await self._on_resolve(provider=provider, instance=provided)
        return provided

    @overload
    async def execute(
        self,
        function: Callable[..., Coroutine[Any, Any, _T]],
        dependencies: Iterable[Dependency[object]],
        *args: Any,
        **kwargs: Any,
    ) -> _T: ...

    @overload
    async def execute(
        self,
        function: Callable[..., _T],
        dependencies: Iterable[Dependency[object]],
        *args: Any,
        **kwargs: Any,
    ) -> _T: ...

    async def execute(
        self,
        function: Callable[..., Coroutine[Any, Any, _T] | _T],
        dependencies: Iterable[Dependency[object]],
        *args: Any,
        **kwargs: Any,
    ) -> _T:
        resolved = {
            dependency.name: await self._resolve(  # type: ignore[call-overload]
                type_=dependency.inner_type,
                is_iterable=dependency.is_iterable,
            )
            for dependency in dependencies
            if dependency.name not in kwargs
        }

        if inspect.iscoroutinefunction(function):
            return await function(*args, **kwargs, **resolved)
        return function(*args, **kwargs, **resolved)  # type: ignore[return-value]

    async def _on_resolve(self, provider: Provider[T], instance: T) -> None:
        for extension in self._extensions:
            if isinstance(extension, OnResolveExtension):
                await extension.on_resolve(self, provider, instance)

    async def __aenter__(self) -> Self:
        self._token = context_var.set(self)
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if self._closed:
            return

        await self._store.__aexit__(exc_type, exc_val, exc_tb)
        context_var.reset(self._token)  # type: ignore[arg-type]
        self._closed = True


class SyncInjectionContext(_BaseInjectionContext[SyncContextExtension]):
    def resolve(self, type_: type[_T]) -> _T:
        return self._resolve(type_, is_iterable=False)

    def resolve_iterable(self, type_: type[_T]) -> list[_T]:
        return self._resolve(type_, is_iterable=True)

    @overload
    def _resolve(
        self,
        type_: type[_T],
        *,
        is_iterable: Literal[False],
    ) -> _T: ...

    @overload
    def _resolve(
        self,
        type_: type[_T],
        *,
        is_iterable: Literal[True],
    ) -> list[_T]: ...

    def _resolve(
        self,
        type_: type[_T],
        *,
        is_iterable: bool,
    ) -> _T | list[_T]:
        providers = self._get_providers(type_)
        if not is_iterable:
            return self._resolve_provider(providers[-1])
        return [self._resolve_provider(provider) for provider in providers]

    def _resolve_provider(
        self,
        provider: Provider[_T],
    ) -> _T:
        store = self._get_store(provider.lifetime)
        if (cached := store.get(provider)) is not NotInCache.sentinel:
            return cached

        provider_dependencies = provider.collect_dependencies(
            context=self._container.type_context
        )
        dependencies_map = get_generic_parameter_map(
            provider.type_,  # type: ignore[arg-type]
            provider_dependencies,
        )
        dependencies = {
            dependency.name: self._resolve(  # type: ignore[call-overload]
                type_=dependencies_map.get(
                    dependency.name,
                    dependency.inner_type,
                ),
                is_iterable=dependency.is_iterable,
            )
            for dependency in provider_dependencies
        }

        if provider.lifetime is DependencyLifetime.singleton:
            with store.sync_lock(provider) as should_provide:
                if should_provide:
                    return self._provide_and_store(
                        provider, store, dependencies
                    )
                return store.get(provider)  # type: ignore[return-value] # pragma: no cover

        return self._provide_and_store(provider, store, dependencies)

    def _provide_and_store(
        self,
        provider: Provider[_T],
        store: InstanceStore,
        dependencies: Mapping[str, object],
    ) -> _T:
        provided = provider.provide_sync(dependencies)
        if provider.is_generator:
            provided = store.enter_sync_context(provided)
        store.add(provider, provided)
        self._on_resolve(provider=provider, instance=provided)
        return provided

    def execute(
        self,
        function: Callable[..., _T],
        dependencies: Iterable[Dependency[object]],
        *args: Any,
        **kwargs: Any,
    ) -> _T:
        resolved = {
            dependency.name: self._resolve(  # type: ignore[call-overload]
                type_=dependency.inner_type,
                is_iterable=dependency.is_iterable,
            )
            for dependency in dependencies
            if dependency.name not in kwargs
        }
        return function(*args, **kwargs, **resolved)

    def _on_resolve(self, provider: Provider[T], instance: T) -> None:
        for extension in self._extensions:
            if isinstance(extension, SyncOnResolveExtension):
                extension.on_resolve_sync(self, provider, instance)

    def __enter__(self) -> Self:
        self._token = context_var.set(self)
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if self._closed:  # pragma: no cover
            return

        self._store.__exit__(exc_type, exc_val, exc_tb)
        context_var.reset(self._token)  # type: ignore[arg-type]
        self._closed = True



from typing import Annotated

from aiogram import Router

from aioinject import Container, Inject, Object
from aioinject.ext.aiogram import AioInjectMiddleware, inject


_NUMBER = 42


async def test_middleware() -> None:
    container = Container()
    container.register(Object(_NUMBER))

    middleware = AioInjectMiddleware(container=container)
    event_ = object()
    data_ = object()

    @inject
    async def handler(
        event: object,
        data: object,
        number: Annotated[int, Inject],
    ) -> None:
        assert event is event_
        assert data is data_
        assert number == _NUMBER

    await middleware(handler=handler, event=event_, data=data_)  # type: ignore[arg-type]


def test_add_to_router() -> None:
    middleware = AioInjectMiddleware(container=Container())

    router = Router()
    middleware.add_to_router(router=router)

    for observer in router.observers.values():
        assert observer.outer_middleware[0] is middleware



import contextlib
from collections.abc import AsyncIterator, Iterator

import pytest
from pydantic_settings import BaseSettings

from aioinject import Object, Scoped, Singleton, providers
from aioinject.containers import Container
from aioinject.context import InjectionContext


class _AbstractService:
    pass


class _ServiceA(_AbstractService):
    pass


class _ServiceB(_AbstractService):
    pass


@pytest.fixture
def container() -> Container:
    return Container()


def test_can_init(container: Container) -> None:
    assert container


def test_can_retrieve_context(container: Container) -> None:
    ctx = container.context()
    assert isinstance(ctx, InjectionContext)


def test_can_register_single(container: Container) -> None:
    provider = providers.Scoped(_ServiceA)
    container.register(provider)

    expected = {_ServiceA: [provider]}
    assert container.providers == expected


def test_can_register_batch(container: Container) -> None:
    provider1 = providers.Scoped(_ServiceA)
    provider2 = providers.Scoped(_ServiceB)
    container.register(provider1, provider2)
    excepted = {_ServiceA: [provider1], _ServiceB: [provider2]}
    assert container.providers == excepted


async def test_register_unhashable_implementation(
    container: Container,
) -> None:
    class ExampleSettings(BaseSettings):
        value: list[str] = []

    container.register(Object([], type_=list[int]))
    container.register(Object(ExampleSettings(), type_=ExampleSettings))


def test_cant_register_multiple_providers_for_same_type(
    container: Container,
) -> None:
    container.register(Scoped(int))

    with pytest.raises(
        ValueError,
        match="^Provider for type <class 'int'> with same implementation already registered$",
    ):
        container.register(Scoped(int))


def test_can_try_register(container: Container) -> None:
    def same_impl() -> _ServiceA:
        return _ServiceA()

    provider = providers.Scoped(same_impl, _ServiceA)
    container.register(provider)

    expected = {_ServiceA: [provider]}
    assert container.providers == expected

    container.try_register(providers.Scoped(same_impl, _ServiceA))
    assert container.providers == expected


def test_can_retrieve_single_provider(container: Container) -> None:
    int_provider = providers.Scoped(int)
    container.register(int_provider)
    assert container.get_provider(int)


def test_can_retrieve_multiple_providers(container: Container) -> None:
    int_providers = [
        providers.Scoped(lambda: 1, int),
        providers.Scoped(lambda: 2, int),
    ]
    container.register(*int_providers)
    assert len(container.get_providers(int)) == len(int_providers)


def test_missing_provider() -> None:
    container = Container()
    with pytest.raises(ValueError) as exc_info:  # noqa: PT011
        assert container.get_provider(_ServiceA)

    msg = f"Providers for type {_ServiceA.__qualname__} not found"
    assert str(exc_info.value) == msg


async def test_should_close_singletons() -> None:
    shutdown = False

    @contextlib.asynccontextmanager
    async def dependency() -> AsyncIterator[int]:
        nonlocal shutdown

        yield 42
        shutdown = True

    container = Container()
    container.register(Singleton(dependency))
    async with container:
        for _ in range(2):
            async with container.context() as ctx:
                assert await ctx.resolve(int) == 42  # noqa: PLR2004

        assert shutdown is False
    assert shutdown is True


def test_should_close_singletons_sync() -> None:
    shutdown = False

    @contextlib.contextmanager
    def dependency() -> Iterator[int]:
        nonlocal shutdown
        yield 42
        shutdown = True

    container = Container()
    container.register(Singleton(dependency))
    with container:
        for _ in range(2):
            with container.sync_context() as ctx:
                assert ctx.resolve(int) == 42  # noqa: PLR2004

        assert shutdown is False
    assert shutdown is True



from datetime import datetime
from typing import NewType

from aioinject import Container, Transient
from aioinject.extensions import OnInitExtension


Now = NewType("Now", datetime)


class TimeExtension(OnInitExtension):
    def on_init(
        self,
        container: Container,
    ) -> None:
        container.register(Transient(datetime.now, Now))


container = Container(extensions=[TimeExtension()])
with container.sync_context() as ctx:
    print(ctx.resolve(Now))



import contextlib
from collections import defaultdict
from collections.abc import Iterator, Sequence
from contextlib import AsyncExitStack
from types import TracebackType
from typing import Any

from typing_extensions import Self

from aioinject import _types
from aioinject._store import SingletonStore
from aioinject._types import T
from aioinject.context import InjectionContext, SyncInjectionContext
from aioinject.extensions import (
    ContextExtension,
    Extension,
    LifespanExtension,
    OnInitExtension,
    SyncContextExtension,
)
from aioinject.providers import Provider


class Container:
    def __init__(self, extensions: Sequence[Extension] | None = None) -> None:
        self._exit_stack = AsyncExitStack()
        self._singletons = SingletonStore(exit_stack=self._exit_stack)

        self.providers: _types.Providers[Any] = defaultdict(list)
        self.type_context: dict[str, type[Any]] = {}
        self.extensions = extensions or []
        self._init_extensions(self.extensions)

    def _init_extensions(self, extensions: Sequence[Extension]) -> None:
        for extension in extensions:
            if isinstance(extension, OnInitExtension):
                extension.on_init(self)

    def register(self, *providers: Provider[Any]) -> None:
        for provider in providers:
            self._register(provider)

    def try_register(self, *providers: Provider[Any]) -> None:
        for provider in providers:
            with contextlib.suppress(ValueError):
                self._register(provider)

    def _register(self, provider: Provider[Any]) -> None:
        if any(
            provider.impl == existing_provider.impl
            for existing_provider in self.providers.get(provider.type_, [])
        ):
            msg = (
                f"Provider for type {provider.type_} with same "
                f"implementation already registered"
            )
            raise ValueError(msg)

        self.providers[provider.type_].append(provider)

        class_name = getattr(provider.type_, "__name__", None)
        if class_name and class_name not in self.type_context:
            self.type_context[class_name] = provider.type_

    def get_provider(self, type_: type[T]) -> Provider[T]:
        return self.get_providers(type_)[0]

    def get_providers(self, type_: type[T]) -> list[Provider[T]]:
        if providers := self.providers[type_]:
            return providers

        err_msg = f"Providers for type {type_.__qualname__} not found"
        raise ValueError(err_msg)

    def context(
        self,
        extensions: Sequence[ContextExtension] = (),
    ) -> InjectionContext:
        return InjectionContext(
            container=self,
            singletons=self._singletons,
            extensions=extensions,
        )

    def sync_context(
        self,
        extensions: Sequence[SyncContextExtension] = (),
    ) -> SyncInjectionContext:
        return SyncInjectionContext(
            container=self,
            singletons=self._singletons,
            extensions=extensions,
        )

    @contextlib.contextmanager
    def override(self, *providers: Provider[Any]) -> Iterator[None]:
        previous = {
            provider.type_: self.providers.get(provider.type_, None)
            for provider in providers
        }
        overridden = defaultdict(
            list,
            {provider.type_: [provider] for provider in providers},
        )

        self.providers.update(overridden)

        try:
            yield
        finally:
            for provider in providers:
                del self.providers[provider.type_]
                if (prev := previous[provider.type_]) is not None:
                    self.providers[provider.type_] = prev

    async def __aenter__(self) -> Self:
        for extension in self.extensions:
            if isinstance(extension, LifespanExtension):
                await self._exit_stack.enter_async_context(
                    extension.lifespan(self),
                )
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self._singletons.__aexit__(exc_type, exc_val, exc_tb)

    async def aclose(self) -> None:
        await self.__aexit__(None, None, None)  # pragma: no cover

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self._singletons.__exit__(exc_type, exc_val, exc_tb)

    def close(self) -> None:
        self.__exit__(None, None, None)  # pragma: no cover



import contextlib
from collections.abc import Iterator

import aioinject
from aioinject import Singleton


@contextlib.contextmanager
def dependency() -> Iterator[int]:
    print("Startup")
    yield 42
    print("Shutdown")


container = aioinject.Container()
container.register(Singleton(dependency))

with container:
    with container.sync_context() as ctx:
        print(ctx.resolve(int))  # Startup, 42
    print("Context is closed")
# Shutdown



import contextlib
import uuid

import httpx
import pytest

import aioinject
from aioinject import Provider, Scoped, Singleton, Transient
from tests.ext.utils import ExceptionPropagation, PropagatedError


async def test_function_route(
    http_client: httpx.AsyncClient,
    provided_value: int,
) -> None:
    response = await http_client.get("/function-route")
    assert response.status_code == httpx.codes.OK.value
    assert response.json() == {"value": provided_value}


async def test_function_route_override(
    http_client: httpx.AsyncClient,
    container: aioinject.Container,
) -> None:
    expected = str(uuid.uuid4())
    with container.override(aioinject.Object(expected, type_=int)):
        response = await http_client.get("/function-route")
    assert response.status_code == httpx.codes.OK.value
    assert response.json() == {"value": expected}


@pytest.mark.parametrize(
    ("provider_type", "should_propagate"),
    [
        (Singleton, False),
        (Scoped, True),
        (Transient, True),
    ],
)
async def test_should_propagate_exceptions(
    http_client: httpx.AsyncClient,
    container: aioinject.Container,
    provider_type: type[Provider[int]],
    should_propagate: bool,
) -> None:
    propagation = ExceptionPropagation()

    with (
        container.override(provider_type(propagation.dependency)),  # type: ignore[call-arg]
        contextlib.suppress(PropagatedError),
    ):
        await http_client.get("/raise-exception")

    if should_propagate:
        assert isinstance(propagation.exc, PropagatedError)
    else:
        assert propagation.exc is None



from __future__ import annotations

import functools
import types
import typing as t
from types import GenericAlias
from typing import TYPE_CHECKING, Any, TypeGuard

from aioinject._utils import is_iterable_generic_collection


if TYPE_CHECKING:
    from aioinject.providers import Dependency


def _is_generic_alias(type_: Any) -> TypeGuard[GenericAlias]:
    return isinstance(
        type_,
        types.GenericAlias | t._GenericAlias,  # type: ignore[attr-defined] # noqa: SLF001
    ) and not is_iterable_generic_collection(type_)


def _get_orig_bases(type_: type) -> tuple[type, ...] | None:
    return getattr(type_, "__orig_bases__", None)


def _get_generic_arguments(type_: Any) -> list[t.TypeVar] | None:
    """
    Returns generic arguments of given class, e.g. Class[T] would return [~T]
    """
    if _is_generic_alias(type_):
        args = t.get_args(type_)
        return [arg for arg in args if isinstance(arg, t.TypeVar)]
    return None


@functools.cache
def _get_generic_args_map(type_: type[object]) -> dict[str, type[object]]:
    if _is_generic_alias(type_):
        params: dict[str, Any] = {
            param.__name__: param
            for param in type_.__origin__.__parameters__  # type: ignore[attr-defined]
        }
        return dict(zip(params, type_.__args__, strict=True))

    args_map = {}
    if orig_bases := _get_orig_bases(type_):
        # find the generic parent
        for base in orig_bases:
            if _is_generic_alias(base):  # noqa: SIM102
                if params := {
                    param.__name__: param
                    for param in getattr(base.__origin__, "__parameters__", ())
                }:
                    args_map.update(
                        dict(zip(params, base.__args__, strict=True)),
                    )
    return args_map


@functools.cache
def get_generic_parameter_map(
    provided_type: type[object],
    dependencies: tuple[Dependency[Any], ...],
) -> dict[str, type[object]]:
    args_map = _get_generic_args_map(provided_type)  # type: ignore[arg-type]
    result = {}
    for dependency in dependencies:
        inner_type = dependency.inner_type
        if args_map and (
            generic_arguments := _get_generic_arguments(inner_type)
        ):
            # This is a generic type, we need to resolve the type arguments
            # and pass them to the provider.
            resolved_args = tuple(
                args_map[arg.__name__] for arg in generic_arguments
            )
            #  We can use `[]` when we drop support for 3.10
            result[dependency.name] = inner_type[resolved_args]
    return result



import abc
from collections.abc import Sequence
from typing import Annotated, NewType

import pytest

from aioinject import Container, Inject, Object, Scoped, inject, providers
from aioinject.context import container_var
from aioinject.decorators import InjectMethod
from aioinject.markers import Injected


class _Session:
    pass


class _Service:
    def __init__(self, session: _Session) -> None:
        self.session = session


class _Interface:
    def __init__(self, session: _Session) -> None:
        self.session = session


@pytest.fixture
def container() -> Container:
    container = Container()
    container.register(providers.Scoped(_Session))
    container.register(providers.Scoped(_Service))
    return container


@inject
def _injectee(
    test: Annotated[_Session, Inject],
    test_no_cache: Annotated[_Session, Inject()],
) -> tuple[_Session, _Session]:
    return test, test_no_cache


def test_would_fail_without_active_context() -> None:
    with pytest.raises(LookupError):
        _injectee()  # type: ignore[call-arg]


def test_would_not_inject_without_inject_marker(container: Container) -> None:
    @inject
    def injectee(session: _Session) -> _Session:
        return session

    with container.sync_context(), pytest.raises(TypeError):
        injectee()  # type: ignore[call-arg]


def test_simple_inject(container: Container) -> None:
    with container.sync_context():
        session, *_ = _injectee()  # type: ignore[call-arg]
        assert isinstance(session, _Session)


def test_simple_inject_with_injected(container: Container) -> None:
    @inject
    def injectee(
        test: Injected[_Session],
        test_no_cache: Injected[_Session],
    ) -> tuple[_Session, _Session]:
        return test, test_no_cache

    with container.sync_context():
        session, *_ = _injectee()  # type: ignore[call-arg]
        assert isinstance(session, _Session)


async def test_simple_service(container: Container) -> None:
    with container.sync_context() as ctx:
        session = ctx.resolve(_Session)
        assert isinstance(session, _Session)

    async with container.context() as ctx:
        session = await ctx.resolve(_Session)
        assert isinstance(session, _Session)


async def test_retrieve_service_with_dependencies(
    container: Container,
) -> None:
    with container.sync_context() as ctx:
        service = ctx.resolve(_Service)
        assert isinstance(service, _Service)
        assert isinstance(service.session, _Session)

    async with container.context() as ctx:
        service = await ctx.resolve(_Service)
        assert isinstance(service, _Service)
        assert isinstance(service.session, _Session)


async def test_inject_using_container(
    container: Container,
) -> None:
    @inject(inject_method=InjectMethod.container)
    async def injectee(service: Annotated[_Service, Inject]) -> _Service:
        return service

    token = container_var.set(container)
    # This is fine
    coro = injectee()  # type: ignore[call-arg]
    assert isinstance(await coro, _Service)
    container_var.reset(token)


async def test_new_type() -> None:
    A = NewType("A", int)
    B = NewType("B", int)

    class Service:
        def __init__(self, a: A, b: B) -> None:
            self.a = a
            self.b = b

    container = Container()
    container.register(Object(1, type_=A))
    container.register(Object(2, type_=B))
    container.register(Scoped(Service))

    async with container.context() as ctx:
        service = await ctx.resolve(Service)
        assert service.a == 1
        assert service.b == 2  # noqa: PLR2004


async def test_iterable_provider() -> None:
    class ILogger(abc.ABC):
        @abc.abstractmethod
        def log(self, msg: str) -> None: ...

    class ConsoleLogger(ILogger):
        def log(self, msg: str) -> None:
            print(msg)  # noqa: T201

    class FileLogger(ILogger):
        def log(self, msg: str) -> None:
            pass

    class Service:
        def __init__(
            self,
            all_loggers: Sequence[ILogger],
            actual_logger: ILogger,
        ) -> None:
            self.all_loggers = all_loggers
            self.actual_logger = actual_logger

    container = Container()
    container.register(Scoped(ConsoleLogger, type_=ILogger))
    container.register(Scoped(FileLogger, type_=ILogger))
    container.register(Scoped(Service))

    async with container.context() as ctx:
        service = await ctx.resolve(Service)
        assert isinstance(service.all_loggers[0], ConsoleLogger)
        assert isinstance(service.all_loggers[1], FileLogger)
        assert isinstance(service.actual_logger, FileLogger)

        loggers = await ctx.resolve_iterable(ILogger)  # type: ignore[type-abstract]
        assert len(loggers) == 2  # noqa: PLR2004

    with container.sync_context() as sync_ctx:
        loggers = sync_ctx.resolve_iterable(ILogger)  # type: ignore[type-abstract]
        assert len(loggers) == 2  # noqa: PLR2004



import functools
from collections.abc import Callable
from typing import ParamSpec, TypeVar


T = TypeVar("T")
P = ParamSpec("P")


def dummy_decorator(func: Callable[P, T]) -> Callable[P, T]:
    @functools.wraps(func)
    def decorator(*args: P.args, **kwargs: P.kwargs) -> T:
        return func(*args, **kwargs)

    return decorator



from typing import TYPE_CHECKING, ParamSpec, TypeAlias, TypeVar, Union


if TYPE_CHECKING:
    from aioinject import InjectionContext, Provider, SyncInjectionContext

P = ParamSpec("P")
T = TypeVar("T")
Providers: TypeAlias = dict[type[T], list["Provider[T]"]]
AnyCtx: TypeAlias = Union["InjectionContext", "SyncInjectionContext"]



import asyncio

from aiogram import Bot, Dispatcher, Router
from aiogram.filters import Command
from aiogram.types import Message
from benchmark.container import create_container

from aioinject import Injected, Object
from aioinject.ext.aiogram import AioInjectMiddleware, inject


async def main() -> None:
    dispatcher = Dispatcher()

    container = create_container()
    container.register(Object(42))

    router = Router()

    @router.message(
        Command(commands=["start"]),
    )
    @inject
    async def start(
        message: Message,
        value: Injected[int],
    ) -> None:
        await message.reply(f"Injected value is {value}")

    middleware = AioInjectMiddleware(container)
    middleware.add_to_router(router)

    dispatcher.include_router(router)

    async with (
        container,
        Bot(token="token-here") as bot,  # noqa: S106
    ):
        await dispatcher.start_polling(bot)


if __name__ == "__main__":
    asyncio.run(main())



from collections.abc import Sequence

from pydantic import PostgresDsn
from pydantic_settings import BaseSettings, SettingsConfigDict

import aioinject


class AppSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="app_")

    version: str
    site_url: str


class DatabaseSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="database_")

    url: PostgresDsn


_settings_classes: Sequence[type[BaseSettings]] = [
    AppSettings,
    DatabaseSettings,
]


def create_container() -> aioinject.Container:
    container = aioinject.Container()

    for settings_cls in _settings_classes:
        # Type is inferred from the instance passed into "Object"
        container.register(aioinject.Object(settings_cls()))

    return container



import uuid
from collections.abc import AsyncIterator
from unittest import mock

import httpx
import pytest
import strawberry
from strawberry.types import ExecutionResult

import aioinject
from tests.ext.conftest import ScopedNode


@pytest.mark.parametrize("resolver_name", ["helloWorld", "helloWorldSync"])
async def test_async_resolver(
    http_client: httpx.AsyncClient,
    provided_value: int,
    resolver_name: str,
) -> None:
    query = """
    query($argument: String!) {
        value: RESOLVER_NAME(argument: $argument)
    }
    """.replace(
        "RESOLVER_NAME",
        resolver_name,
    )
    argument = str(uuid.uuid4())
    response = await http_client.post(
        "",
        json={"query": query, "variables": {"argument": argument}},
    )
    assert response.status_code == httpx.codes.OK.value

    assert response.json() == {
        "data": {"value": f"{argument}-{provided_value}"},
    }


async def test_dataloader(
    http_client: httpx.AsyncClient,
) -> None:
    query = """
    query {
        numbers: dataloader
    }
    """
    str(uuid.uuid4())
    response = await http_client.post("", json={"query": query})
    assert response.status_code == httpx.codes.OK.value

    assert response.json() == {
        "data": {"numbers": list(range(100))},
    }


def ensure_agen(
    gen: object,
) -> AsyncIterator[ExecutionResult]:
    if not hasattr(gen, "__aiter__"):
        msg = f"Expected an async generator, got {gen!r}"
        raise TypeError(msg)
    return gen  # type: ignore[return-value]


async def test_subscription(
    container: aioinject.Container,
    schema: strawberry.Schema,
) -> None:
    subscription = """
    subscription {
        liveBars {
            id
            baz
        }
    }
    """
    generate_node_mock = mock.Mock()
    generate_node_mock.return_value = {
        "id": "5986123d250742a681da1defac165b8e",
    }
    provider = container.get_provider(ScopedNode)

    def wrap_mock() -> ScopedNode:
        return generate_node_mock()

    with container.override(provider, aioinject.Scoped(wrap_mock)):
        results = []
        async for res in ensure_agen(await schema.subscribe(subscription)):
            assert res.data
            assert not res.errors
            results.append(res.data["liveBars"])

        assert results == [
            {
                "id": "5986123d250742a681da1defac165b8e",
                "baz": "baz-5986123d250742a681da1defac165b8e",
            }
            for _ in range(5)
        ]



from types import TracebackType

from typing_extensions import Self

import aioinject
from aioinject import Scoped


class Class:
    def __enter__(self) -> Self:
        print("Startup")
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        print("Shutdown")


container = aioinject.Container()
container.register(Scoped(Class))

with container.sync_context() as ctx:
    print(ctx.resolve(Class))  # <__main__.Class object at ...>



import time
from collections.abc import AsyncIterator
from datetime import timedelta
from typing import Annotated

import httpx
from httpx import ASGITransport
from litestar import Litestar, get
from litestar.di import Provide
from litestar.logging import LoggingConfig

from aioinject import Inject
from aioinject.ext.fastapi import inject
from aioinject.ext.litestar import AioInjectPlugin
from benchmark.benches._common import (
    _get_repository_a,
    _get_repository_b,
    _get_service_a,
    _get_service_b,
    _get_session,
    _get_usecase,
)
from benchmark.container import create_container
from benchmark.dependencies import (
    UseCase,
)
from benchmark.dto import BenchmarkResult


@get("/aioinject")
@inject
async def test_aioinject(use_case: Annotated[UseCase, Inject]) -> int:
    return await use_case.execute()


@get("/litestar")
async def test_litestar(use_case: UseCase) -> int:
    return await use_case.execute()


async def litestar_bench(
    iterations: int,
    endpoint: str,
    *,
    enable_aioinject: bool = False,
) -> AsyncIterator[BenchmarkResult]:
    app = Litestar(
        plugins=(
            [AioInjectPlugin(container=create_container())]
            if enable_aioinject
            else []
        ),
        route_handlers=[test_aioinject, test_litestar],
        dependencies={
            "session": Provide(_get_session),
            "repository_a": Provide(_get_repository_a),
            "repository_b": Provide(_get_repository_b),
            "service_a": Provide(_get_service_a),
            "service_b": Provide(_get_service_b),
            "use_case": Provide(_get_usecase),
        },
        logging_config=LoggingConfig(disable_existing_loggers=True),
    )

    durations = []
    async with httpx.AsyncClient(
        transport=ASGITransport(app),  # type: ignore[arg-type]
        base_url="http://test",
    ) as client:
        for _ in range(iterations):
            start = time.perf_counter()
            response = await client.get(endpoint)
            durations.append(
                timedelta(seconds=time.perf_counter() - start),
            )
            response.raise_for_status()
            assert response.content == b"42"  # noqa: S101

    yield BenchmarkResult(
        name=f"Litestar - {endpoint}",
        durations=durations,
        iterations=iterations,
    )



from collections.abc import AsyncIterator
from typing import Annotated

from fastapi import Depends

from benchmark.dependencies import (
    RepositoryA,
    RepositoryB,
    ServiceA,
    ServiceB,
    Session,
    UseCase,
)


async def _get_session() -> AsyncIterator[Session]:
    yield Session()


SessionDepends = Annotated[Session, Depends(_get_session)]


async def _get_repository_a(session: SessionDepends) -> RepositoryA:
    return RepositoryA(session)


RepositoryADepends = Annotated[RepositoryA, Depends(_get_repository_a)]


async def _get_repository_b(session: SessionDepends) -> RepositoryB:
    return RepositoryB(session)


RepositoryBDepends = Annotated[RepositoryB, Depends(_get_repository_b)]


async def _get_service_a(repository_a: RepositoryADepends) -> ServiceA:
    return ServiceA(repository=repository_a)


ServiceADepends = Annotated[ServiceA, Depends(_get_service_a)]


async def _get_service_b(repository_b: RepositoryBDepends) -> ServiceB:
    return ServiceB(repository=repository_b)


ServiceBDepends = Annotated[ServiceB, Depends(_get_service_b)]


async def _get_usecase(
    service_a: ServiceADepends,
    service_b: ServiceBDepends,
) -> UseCase:
    return UseCase(
        service_a=service_a,
        service_b=service_b,
    )


UseCaseDepends = Annotated[UseCase, Depends(_get_usecase)]



import aioinject
from aioinject import Object, Scoped


async def test_ok() -> None:
    container = aioinject.Container()
    async with container.context() as ctx:
        ctx.register(Scoped(int))
        assert await ctx.resolve(int) == 0


async def test_should_override_container_provider() -> None:
    value = 42
    container = aioinject.Container()
    container.register(Object(value))

    async with container.context() as ctx:
        assert await ctx.resolve(int) == value

    async with container.context() as ctx:
        ctx.register(Scoped(int))
        assert await ctx.resolve(int) == 0



import contextlib
from collections.abc import Iterator

import aioinject
from aioinject import Scoped


@contextlib.contextmanager
def dependency() -> Iterator[int]:
    print("Startup")
    yield 42
    print("Shutdown")


container = aioinject.Container()
container.register(Scoped(dependency))

with container.sync_context() as ctx:
    print(ctx.resolve(int))  # Startup, 42
# Shutdown



from typing import Any

import aioinject
from aioinject import InjectionContext, Provider, SyncInjectionContext
from aioinject.extensions import OnResolveExtension, SyncOnResolveExtension


class _TestExtension(SyncOnResolveExtension, OnResolveExtension):
    async def on_resolve(
        self,
        context: InjectionContext,
        provider: Provider[Any],
        instance: Any,
    ) -> None: ...

    def on_resolve_sync(
        self,
        context: SyncInjectionContext,
        provider: Provider[Any],
        instance: Any,
    ) -> None: ...


async def _pass() -> None:
    # Same extension should be compatible with both contexts
    container = aioinject.Container()
    container.context(extensions=[_TestExtension()])
    container.sync_context(extensions=[_TestExtension()])



from __future__ import annotations

from dataclasses import dataclass

from aioinject.containers import Container
from aioinject.providers import Singleton


@dataclass
class B:
    a: A


cont = Container()


def get_b(a: A) -> B:
    return B(a)


cont.register(Singleton(get_b))


class A: ...


cont.register(Singleton(A))
with cont.sync_context() as ctx:
    a = ctx.resolve(A)
    b = ctx.resolve(B)
    assert b.a is a



from __future__ import annotations

from typing import TYPE_CHECKING, Protocol, runtime_checkable


if TYPE_CHECKING:
    from contextlib import AbstractAsyncContextManager

    from aioinject import (
        Container,
        InjectionContext,
        Provider,
        SyncInjectionContext,
    )
    from aioinject._types import T


@runtime_checkable
class LifespanExtension(Protocol):
    def lifespan(
        self,
        container: Container,
    ) -> AbstractAsyncContextManager[None]: ...


@runtime_checkable
class OnInitExtension(Protocol):
    def on_init(
        self,
        container: Container,
    ) -> None: ...


@runtime_checkable
class OnResolveExtension(Protocol):
    async def on_resolve(
        self,
        context: InjectionContext,
        provider: Provider[T],
        instance: T,
    ) -> None: ...


@runtime_checkable
class SyncOnResolveExtension(Protocol):
    def on_resolve_sync(
        self,
        context: SyncInjectionContext,
        provider: Provider[T],
        instance: T,
    ) -> None: ...


Extension = LifespanExtension | OnInitExtension
ContextExtension = OnResolveExtension
SyncContextExtension = SyncOnResolveExtension



from __future__ import annotations

from aioinject.containers import Container
from aioinject.providers import Singleton


cont = Container()


# notice that C is not defined yet
def get_c() -> C:
    return C()


cont.register(Singleton(get_c))


class C: ...
